@@ -5,7 +5,7 @@
  * generic text renderer.
  *
  * Copyright (C) 2013-2019 Tom Hacohen <tom at stosb dot com>
- * Copyright (C) 2018 Wu Yongwei <wuyongwei at gmail dot com>
+ * Copyright (C) 2018-2024 Wu Yongwei <wuyongwei at gmail dot com>
  *
  * This software is provided 'as-is', without any express or implied
  * warranty.  In no event will the author be held liable for any damages
@@ -31,9 +31,9 @@
  * Unicode 6.0.0:
  *      <URL:http://www.unicode.org/reports/tr29/tr29-17.html>
  *
- * This library has been updated according to Revision 37, for
- * Unicode 13.0.0:
- *      <URL:http://www.unicode.org/reports/tr29/tr29-37.html>
+ * This library has been updated according to Revision 43, for
+ * Unicode 15.1.0:
+ *      <URL:http://www.unicode.org/reports/tr29/tr29-43.html>
  *
  * The Unicode Terms of Use are available at
  *      <URL:http://www.unicode.org/copyright.html>
@@ -72,33 +72,15 @@ void init_wordbreak(void)
  * @param len  size of the wbp array in number of items
  * @return     the word breaking class if found; \c WBP_Any otherwise
  */
-static enum WordBreakClass get_char_wb_class(
-        utf32_t ch,
-        const struct WordBreakProperties *wbp,
-        size_t len)
+static enum WordBreakClass get_char_wb_class(utf32_t ch)
 {
-    int min = 0;
-    int max = len - 1;
-    int mid;
-
-    do
+    const struct WordBreakProperties *result_ptr =
+        ub_bsearch(ch, wb_prop_default, ARRAY_LEN(wb_prop_default) - 1,
+                   sizeof(struct WordBreakProperties));
+    if (result_ptr)
     {
-        mid = (min + max) / 2;
-
-        if (ch < wbp[mid].start)
-        {
-            max = mid - 1;
-        }
-        else if (ch > wbp[mid].end)
-        {
-            min = mid + 1;
-        }
-        else
-        {
-            return wbp[mid].prop;
-        }
+        return result_ptr->prop;
     }
-    while (min <= max);
 
     return WBP_Any;
 }
@@ -197,8 +179,7 @@ static void set_wordbreaks(
     while (ch != EOS)
     {
         enum WordBreakClass wbcCur;
-        wbcCur = get_char_wb_class(ch, wb_prop_default,
-                                   ARRAY_LEN(wb_prop_default));
+        wbcCur = get_char_wb_class(ch);
 
         switch (wbcCur)
         {
